use std::{borrow::Cow, clone::Clone, net::ToSocketAddrs};

use once_cell::sync::{Lazy, OnceCell};
use prometheus::{register_int_counter, register_int_counter_vec, register_int_gauge_vec, IntCounter, IntGaugeVec};
use prometheus_static_metric::make_static_metric;
use rustracing_jaeger::{span::SpanReceiver, Span, Tracer};
use tiny_http::{Response, ResponseBox};

use crate::AppError;

#[cfg(debug_assertions)]
static SERVICE_NAME: &str = "fileserver-testing";

#[cfg(not(debug_assertions))]
static SERVICE_NAME: &str = "fileserver";

pub static TRACER_QUIT: std::sync::atomic::AtomicBool = std::sync::atomic::AtomicBool::new(false);
static TRACER_INNER: Lazy<(Tracer, SpanReceiver)> = Lazy::new(|| Tracer::new(rustracing::sampler::AllSampler));
pub static TRACER: Lazy<Tracer> = Lazy::new(|| TRACER_INNER.0.clone());
pub static TRACER_THREAD: OnceCell<std::thread::JoinHandle<()>> = OnceCell::new();

make_static_metric! {
    struct CacheMetrics: IntCounter {
        "cache" => {
            node,
            children
        },
        "type" => {
            hit,
            miss
        }
    }
}

pub static NODES: Lazy<IntGaugeVec> =
    Lazy::new(|| register_int_gauge_vec!("nodes", "All nodes by user, type", &["type", "user"]).unwrap());
pub static DISK_USAGE: Lazy<IntGaugeVec> =
    Lazy::new(|| register_int_gauge_vec!("disk_usage", "Disk usage by user", &["user"]).unwrap());
pub static REQUEST: Lazy<IntCounter> = Lazy::new(|| register_int_counter!("request", "Count of requests").unwrap());
pub static CACHE: Lazy<CacheMetrics> = Lazy::new(|| {
    CacheMetrics::from(&register_int_counter_vec!("cache", "Node cache hits/misses", &["cache", "type"]).unwrap())
});

pub fn init(mut db: crate::db::DBConnection) {
    TRACER_THREAD
        .set(std::thread::spawn(|| {
            let recv = TRACER_INNER.1.clone();
            let mut reporter = rustracing_jaeger::reporter::JaegerCompactReporter::new(SERVICE_NAME).unwrap();
            reporter.set_agent_addr("grafana.lan:6831".to_socket_addrs().unwrap().next().unwrap());
            reporter.set_reporter_addr("0.0.0.0:0".parse().unwrap()).unwrap();
            while let Ok(span) = recv.recv() {
                if TRACER_QUIT.load(std::sync::atomic::Ordering::Relaxed) {
                    break;
                }
                reporter.report(&[span]).unwrap();
            }
        }))
        .unwrap();

    let nodes = Lazy::force(&NODES);
    let disk_usage = Lazy::force(&DISK_USAGE);
    Lazy::force(&REQUEST);
    Lazy::force(&CACHE);
    db.get_all_nodes().iter().for_each(|n| {
        let owner = n.owner_id.to_string();
        if n.is_file {
            nodes.with_label_values(&["file", owner.as_str()]).inc();
            disk_usage.with_label_values(&[owner.as_str()]).add(n.size.unwrap_or(0));
        } else {
            nodes.with_label_values(&["folder", owner.as_str()]).inc();
        }
    });
}

pub fn get_metrics() -> Result<ResponseBox, AppError> {
    let metric = prometheus::gather();
    Ok(Response::from_string(prometheus::TextEncoder::new().encode_to_string(&metric).unwrap()).boxed())
}

pub fn span<N>(name: N, span: &Span) -> Span
where N: Into<Cow<'static, str>> {
    TRACER.span(name).child_of(span.context().unwrap()).start()
}
